from transonic import boost, Array
import numba

import numpy as np

Image = Array[np.float64, "2d", "C"]


def laplace_numpy(image: Image):
    """Laplace operator in NumPy for 2D images."""
    laplacian = (
        image[:-2, 1:-1]
        + image[2:, 1:-1]
        + image[1:-1, :-2]
        + image[1:-1, 2:]
        - 4 * image[1:-1, 1:-1]
    )
    thresh = np.abs(laplacian) > 0.05
    return thresh


def laplace_loops(image: Image):
    """Laplace operator for 2D images."""
    h = image.shape[0]
    w = image.shape[1]
    laplacian = np.empty((h - 2, w - 2), np.uint8)
    for i in range(1, h - 1):
        for j in range(1, w - 1):
            laplacian[i - 1, j - 1] = (
                np.abs(
                    image[i - 1, j]
                    + image[i + 1, j]
                    + image[i, j - 1]
                    + image[i, j + 1]
                    - 4 * image[i, j]
                )
                > 0.05
            )
    return laplacian


laplace_transonic_pythran = boost(backend="pythran")(laplace_numpy)
laplace_transonic_cython = boost(backend="cython")(laplace_numpy)
laplace_transonic_numba = boost(backend="numba")(laplace_numpy)
laplace_transonic_python = boost(backend="python")(laplace_numpy)
laplace_numba = numba.njit(laplace_numpy)


laplace_loops_transonic_pythran = boost(backend="pythran")(laplace_loops)
laplace_loops_transonic_python = boost(backend="python")(laplace_loops)
laplace_loops_transonic_numba = boost(backend="numba")(laplace_loops)
laplace_loops_numba = numba.njit(laplace_loops)


# For Cython, we need to add more type annotations

@boost(backend="cython", boundscheck=False, wraparound=False)
def laplace_loops_transonic_cython(image: Image):
    """Laplace operator for 2D images."""
    i: int
    j: int
    h: int = image.shape[0]
    w: int = image.shape[1]
    laplacian: Array[np.uint8, "2d"] = np.empty((h - 2, w - 2), np.uint8)
    for i in range(1, h - 1):
        for j in range(1, w - 1):
            laplacian[i - 1, j - 1] = (
                abs(
                    image[i - 1, j]
                    + image[i + 1, j]
                    + image[i, j - 1]
                    + image[i, j + 1]
                    - 4 * image[i, j]
                )
                > 0.05
            )
    return laplacian


if __name__ == "__main__":

    from skimage.data import astronaut
    from skimage.color import rgb2gray

    image = astronaut()
    image = rgb2gray(image)

    # call these functions to warm them
    laplace_transonic_numba(image)
    laplace_loops_transonic_numba(image)
    laplace_numba(image)
    laplace_loops_numba(image)

    from transonic.util import timeit
    from transonic import __version__
    import pythran

    loc = locals()

    def bench(call, norm=None):
        ret = result = timeit(call, globals=loc)
        if norm is None:
            norm = result
        result /= norm
        print(f"{call.split('(')[0]:33s}: {result:.2f}")
        return ret

    print(
        f"transonic {__version__}\n"
        f"pythran {pythran.__version__}\n"
        f"numba {numba.__version__}\n"
    )

    norm = bench("laplace_transonic_pythran(image)")
    print(f"norm = {norm:.2e} s")
    bench("laplace_loops_transonic_pythran(image)", norm=norm)
    bench("laplace_transonic_cython(image)", norm=norm)
    bench("laplace_loops_transonic_cython(image)", norm=norm)
    bench("laplace_numba(image)", norm=norm)
    bench("laplace_transonic_numba(image)", norm=norm)
    bench("laplace_loops_numba(image)", norm=norm)
    bench("laplace_loops_transonic_numba(image)", norm=norm)
    bench("laplace_numpy(image)", norm=norm)
    bench("laplace_transonic_python(image)", norm=norm)



from distutils.core import setup, Extension
from pathlib import Path

import numpy as np
from Cython.Build import cythonize

path_here = Path(__file__).parent.absolute()
include_dirs = [np.get_include()]

pyx_files = (path_here / "pyx").glob("*.pyx")
extensions = []
for pyx_file in pyx_files:
    name = pyx_file.name.split(".", 1)[0]
    extensions.append(
        Extension("pyx." + name, [str(pyx_file)], include_dirs=include_dirs)
    )

extensions = cythonize(extensions, language_level=3, annotate=True)

setup(
    name="pyx",
    ext_modules=extensions,
    script_args=["build_ext", "--inplace"],
)



import numpy as np

from transonic import boost, Array

A = Array[float, "1d"]


def fxfy(ft: A, fn: A, theta: A):
    sin_theta = np.sin(theta)
    cos_theta = np.cos(theta)
    fx = cos_theta * ft - sin_theta * fn
    fy = sin_theta * ft + cos_theta * fn
    return fx, fy


def fxfy_loops(ft: A, fn: A, theta: A):
    n0 = theta.size
    fx = np.empty_like(ft)
    fy = np.empty_like(fn)
    for index in range(n0):
        sin_theta = np.sin(theta[index])
        cos_theta = np.cos(theta[index])
        fx[index] = cos_theta * ft[index] - sin_theta * fn[index]
        fy[index] = sin_theta * ft[index] + cos_theta * fn[index]
    return fx, fy


fxfy_pythran = boost(backend="pythran")(fxfy)
fxfy_numba = boost(backend="numba")(fxfy)

fxfy_loops_pythran = boost(backend="pythran")(fxfy_loops)
fxfy_loops_numba = boost(backend="numba")(fxfy_loops)


if __name__ == "__main__":

    from transonic.util import print_versions, timeit_verbose

    print_versions()

    theta = np.linspace(0, 2 * np.pi, 10000)
    ft = 2.5 * theta
    fv = 1.5 * theta
    loc = locals()

    out = fxfy(ft, fv, theta)
    out_loops = fxfy_loops(ft, fv, theta)
    assert np.allclose(out, out_loops)

    print()
    norm = timeit_verbose("fxfy(ft, fv, theta)", globals=loc)

    for backend in ("numba", "pythran"):
        timeit_verbose(f"fxfy_{backend}(ft, fv, theta)", globals=loc, norm=norm)
        timeit_verbose(
            f"fxfy_loops_{backend}(ft, fv, theta)", globals=loc, norm=norm
        )




import numpy as np
import matplotlib.pyplot as plt

from transonic import boost, Type, set_backend_for_this_module

set_backend_for_this_module("cython")

T = Type(float, int)

@boost
def func(a: T):
    b = 1
    return a * np.sin(b)



import numpy as np

from transonic import boost, Optional, Array

A = Array[np.uint8, "2d", "memview"]

A1d = Array[np.intp, "1d", "memview"]


@boost(wraparound=False, boundscheck=False, cdivision=True, nonecheck=False)
def _dilate(
    image: A,
    selem: A,
    out: Optional[A] = None,
    shift_x: np.int8 = 0,
    shift_y: np.int8 = 0,
):
    """Return greyscale morphological dilation of an image.

    Morphological dilation sets a pixel at (i,j) to the maximum over all pixels
    in the neighborhood centered at (i,j). Dilation enlarges bright regions
    and shrinks dark regions.

    Parameters
    ----------

    image : ndarray
        Image array.
    selem : ndarray
        The neighborhood expressed as a 2-D array of 1's and 0's.
    out : ndarray
        The array to store the result of the morphology. If None, is
        passed, a new array will be allocated.
    shift_x, shift_y : bool
        shift structuring element about center point. This only affects
        eccentric structuring elements (i.e. selem with even numbered sides).

    Returns
    -------
    dilated : uint8 array
        The result of the morphological dilation.
    """

    rows: np.intp = image.shape[0]
    cols: np.intp = image.shape[1]
    srows: np.intp = selem.shape[0]
    scols: np.intp = selem.shape[1]

    centre_r: np.intp = int(srows / 2) - shift_y
    centre_c: np.intp = int(scols / 2) - shift_x

    image = np.ascontiguousarray(image)
    if out is None:
        out = np.zeros((rows, cols), dtype=np.uint8)

    selem_num: np.intp = np.sum(np.asarray(selem) != 0)
    sr: A1d = np.empty(selem_num, dtype=np.intp)
    sc: A1d = np.empty(selem_num, dtype=np.intp)

    s: np.intp = 0
    r: np.intp
    c: np.intp

    for r in range(srows):
        for c in range(scols):
            if selem[r, c] != 0:
                sr[s] = r - centre_r
                sc[s] = c - centre_c
                s += 1

    local_max: np.uint8
    value: np.uint8
    rr: np.intp
    cc: np.intp

    for r in range(rows):
        for c in range(cols):
            local_max = 0
            for s in range(selem_num):
                rr = r + sr[s]
                cc = c + sc[s]
                if 0 <= rr < rows and 0 <= cc < cols:
                    value = image[rr, cc]
                    if value > local_max:
                        local_max = value

            out[r, c] = local_max

    return np.asarray(out)


@boost(wraparound=False, boundscheck=False, cdivision=True, nonecheck=False)
def _erode(
    image: A,
    selem: A,
    out: Optional[A] = None,
    shift_x: np.int8 = 0,
    shift_y: np.int8 = 0,
):
    """Return greyscale morphological erosion of an image.

    Morphological erosion sets a pixel at (i,j) to the minimum over all pixels
    in the neighborhood centered at (i,j). Erosion shrinks bright regions and
    enlarges dark regions.

    Parameters
    ----------
    image : ndarray
        Image array.
    selem : ndarray
        The neighborhood expressed as a 2-D array of 1's and 0's.
    out : ndarray
        The array to store the result of the morphology. If None is
        passed, a new array will be allocated.
    shift_x, shift_y : bool
        shift structuring element about center point. This only affects
        eccentric structuring elements (i.e. selem with even numbered sides).

    Returns
    -------
    eroded : uint8 array
        The result of the morphological erosion.
    """

    rows: np.intp = image.shape[0]
    cols: np.intp = image.shape[1]
    srows: np.intp = selem.shape[0]
    scols: np.intp = selem.shape[1]

    centre_r: np.intp = int(srows / 2) - shift_y
    centre_c: np.intp = int(scols / 2) - shift_x

    image = np.ascontiguousarray(image)
    if out is None:
        out = np.zeros((rows, cols), dtype=np.uint8)

    selem_num: np.intp = np.sum(np.asarray(selem) != 0)
    sr: A1d = np.empty(selem_num, dtype=np.intp)
    sc: A1d = np.empty(selem_num, dtype=np.intp)

    s: np.intp = 0
    r: np.intp
    c: np.intp

    for r in range(srows):
        for c in range(scols):
            if selem[r, c] != 0:
                sr[s] = r - centre_r
                sc[s] = c - centre_c
                s += 1

    local_max: np.uint8
    value: np.uint8
    rr: np.intp
    cc: np.intp

    for r in range(rows):
        for c in range(cols):
            local_min = 255
            for s in range(selem_num):
                rr = r + sr[s]
                cc = c + sc[s]
                if 0 <= rr < rows and 0 <= cc < cols:
                    value = image[rr, cc]
                    if value < local_min:
                        local_min = value

            out[r, c] = local_min

    return np.asarray(out)



# __protected__ from numba import njit
import numpy as np

# __protected__ @njit(cache=True, fastmath=True)


def __for_method__Transmitter____call__(self_arr, self_freq, inp):
    "My docstring"
    return ((inp * np.exp(((np.arange(len(inp)) * self_freq) * 1j))), self_arr)


__code_new_method__Transmitter____call__ = "\n\ndef new_method(self, inp):\n    return backend_func(self.arr, self.freq, inp)\n\n"
__transonic__ = ("0.4.7",)



#cython: cdivision=True
#cython: nonecheck=False
#cython: wraparound=False
import numpy as np


def moments_hu(double[:, :] nu):
    cdef double[::1] hu = np.zeros((7, ), dtype=np.double)
    cdef double t0 = nu[3, 0] + nu[1, 2]
    cdef double t1 = nu[2, 1] + nu[0, 3]
    cdef double q0 = t0 * t0
    cdef double q1 = t1 * t1
    cdef double n4 = 4 * nu[1, 1]
    cdef double s = nu[2, 0] + nu[0, 2]
    cdef double d = nu[2, 0] - nu[0, 2]
    hu[0] = s
    hu[1] = d * d + n4 * nu[1, 1]
    hu[3] = q0 + q1
    hu[5] = d * (q0 - q1) + n4 * t0 * t1
    t0 *= q0 - 3 * q1
    t1 *= 3 * q0 - q1
    q0 = nu[3, 0]- 3 * nu[1, 2]
    q1 = 3 * nu[2, 1] - nu[0, 3]
    hu[2] = q0 * q0 + q1 * q1
    hu[4] = q0 * t0 + q1 * t1
    hu[6] = q1 * t0 - q0 * t1
    return np.asarray(hu)



import numpy as np
from transonic import boost, Array, Type

A = Array[Type(np.float64, np.complex128), "3d"]
Af = "float[:,:,:]"
A = Af  # issue fused type with Cython


def proj(vx: A, vy: A, vz: A, kx: Af, ky: Af, kz: Af, inv_k_square_nozero: Af):
    tmp = (kx * vx + ky * vy + kz * vz) * inv_k_square_nozero
    vx -= kx * tmp
    vy -= ky * tmp
    vz -= kz * tmp


def proj_loop(
    vx: A, vy: A, vz: A, kx: Af, ky: Af, kz: Af, inv_k_square_nozero: Af
):

    # type annotations only useful for Cython
    n0: int
    n1: int
    n2: int
    i0: int
    i1: int
    i2: int
    tmp: float

    n0, n1, n2 = kx.shape[0], kx.shape[1], kx.shape[2]

    for i0 in range(n0):
        for i1 in range(n1):
            for i2 in range(n2):
                tmp = (
                    kx[i0, i1, i2] * vx[i0, i1, i2]
                    + ky[i0, i1, i2] * vy[i0, i1, i2]
                    + kz[i0, i1, i2] * vz[i0, i1, i2]
                ) * inv_k_square_nozero[i0, i1, i2]

                vx[i0, i1, i2] -= kx[i0, i1, i2] * tmp
                vy[i0, i1, i2] -= ky[i0, i1, i2] * tmp
                vz[i0, i1, i2] -= kz[i0, i1, i2] * tmp


proj_pythran = boost(backend="pythran")(proj)
proj_numba = boost(backend="numba")(proj)
proj_cython = boost(backend="cython")(proj)

proj_loop_pythran = boost(backend="pythran")(proj_loop)
proj_loop_numba = boost(backend="numba")(proj_loop)
proj_loop_cython = boost(backend="cython", boundscheck=False, wraparound=False)(
    proj_loop
)


if __name__ == "__main__":
    from textwrap import dedent

    from transonic.util import print_versions, timeit_verbose

    loc = locals()

    print_versions()

    setup = dedent(
        """
        shape = n0, n1, n2 = 64, 512, 512
        k0 = np.linspace(0, 100, n0)
        k1 = np.linspace(0, 100, n1)
        k2 = np.linspace(0, 100, n2)
        K1, K0, K2 = np.meshgrid(k1, k0, k2, copy=False)
        kz = np.ascontiguousarray(K0)
        ky = np.ascontiguousarray(K1)
        kx = np.ascontiguousarray(K2)

        k_square_nozero = K0 ** 2 + K1 ** 2 + K2 ** 2
        k_square_nozero[0, 0, 0] = 1e-14
        inv_k_square_nozero = 1.0 / k_square_nozero

        vx = np.ones(shape)
        vy = np.ones(shape)
        vz = np.ones(shape)
    """
    )

    print()
    norm = timeit_verbose(
        "proj(vx, vy, vz, kx, ky, kz, inv_k_square_nozero)",
        setup=setup,
        globals=loc,
    )

    for backend in ("cython", "numba", "pythran"):
        timeit_verbose(
            f"proj_{backend}(vx, vy, vz, kx, ky, kz, inv_k_square_nozero)",
            setup=setup,
            globals=loc,
            norm=norm,
        )
        timeit_verbose(
            f"proj_loop_{backend}(vx, vy, vz, kx, ky, kz, inv_k_square_nozero)",
            setup=setup,
            globals=loc,
            norm=norm,
        )



from transonic import boost

T = int


@boost(inline=True)
def add(a: T, b: T) -> T:
    return a + b


@boost
def use_add(n: int = 10000):
    tmp: T = 0
    _: T
    for _ in range(n):
        tmp = add(tmp, 1)
    return tmp


if __name__ == "__main__":

    from transonic.util import timeit

    n = 100000
    assert n == use_add(n)

    print(f"{timeit('use_add(n)', globals=locals()): .2e} s")



import numpy as np

from transonic.util import timeit
from transonic.config import backend_default


def check(functions, arr, columns):
    res0 = functions[0](arr, columns)
    for func in functions[1:]:
        assert np.allclose(res0, func(arr, columns))
    print("Checks passed: results are consistent")


def bench(functions, arr, columns):
    print(backend_default.capitalize())
    for func in functions:
        result = timeit("func(arr, columns)", globals=locals())
        print(f"{func.__name__:20s} {result:.3e} s")
    print()



# cython: language_level=3
import cython


def add(a, b):
    return a + b


def use_add(n):
    result = 1
    with cython.nogil:
        for i in range(n):
            result = add(result, result)
    return result



# __protected__ from numba import njit
# __protected__ @njit(cache=True, fastmath=True)


def add(a, b):
    return a + b


# __protected__ @njit(cache=True, fastmath=True)


def use_add(n=10000):
    tmp = 0
    for _ in range(n):
        tmp = add(tmp, 1)
    return tmp


__transonic__ = ("0.4.7",)



def block0(a, b, n):

    # transonic block (
    #     A a; A1 b;
    #     int n
    # )
    # transonic block (
    #     int[:] a, b;
    #     float n
    # )
    result = ((a**2) + (b.mean() ** 3)) + n

    return result


arguments_blocks = {"block0": ["a", "b", "n"]}

__transonic__ = ("0.3.0.post0",)



import cython

import numpy as np
cimport numpy as np

cpdef rk2_step0(np.ndarray[np.complex128_t, ndim=3] state_spect_n12, np.ndarray[np.complex128_t, ndim=3] state_spect, np.ndarray[np.complex128_t, ndim=3] tendencies_n, np.ndarray[np.float64_t, ndim=2] diss2, cython.float dt)



# __protected__ from numba import njit
# __protected__ @njit(cache=True, fastmath=True)


def func():
    return 1


# __protected__ @njit(cache=True, fastmath=True)


def func2():
    return 1


__transonic__ = ("0.4.7",)



def rk2_step0(state_spect_n12, state_spect, tendencies_n, diss2, dt):

    # transonic block (
    #     complex128[][][] state_spect_n12, state_spect,
    #                      tendencies_n;
    #     float64[][] diss2;
    #     float dt
    # )
    state_spect_n12[:] = (state_spect + dt / 2 * tendencies_n) * diss2


arguments_blocks = {
    "rk2_step0": ["state_spect_n12", "state_spect", "tendencies_n", "diss2", "dt"]
}

__transonic__ = ("0.3.0.post0",)



from transonic import boost


@boost
def f():
    pass


@boost(inline=1)
def f():
    pass


def f1():
    pass


def f2():
    pass


f1_ = boost(f1)
f1_ = boost(inline=1)(f1)


# decor = boost(inline=1)
# f1_ = decor(f1)
# f2_ = decor(f2)



import numpy as np
from __ext__MyClass2__exterior_import_boost import func_import


def __for_method__MyClass2__myfunc(self_attr0, self_attr1, arg):
    return self_attr1 + self_attr0 + np.abs(arg) + func_import()


__code_new_method__MyClass2__myfunc = """

def new_method(self, arg):
    return backend_func(self.attr0, self.attr1, arg)

"""

__transonic__ = ("0.3.0",)



"""
No pyproject.toml file because in some cases, isolated build cannot be used.

"""
import sys
import os
from pathlib import Path

from runpy import run_path

from setuptools.dist import Distribution
from setuptools import setup, find_packages

if sys.version_info[:2] < (3, 6):
    raise RuntimeError("Python version >= 3.6 required.")


def install_setup_requires():
    dist = Distribution()
    # Honor setup.cfg's options.
    dist.parse_config_files(ignore_option_errors=True)
    if dist.setup_requires:
        dist.fetch_build_eggs(dist.setup_requires)


install_setup_requires()

from transonic.dist import make_backend_files, init_transonic_extensions
import numpy as np

here = Path(__file__).parent.absolute()

pack_name = "package_simple"
pack_dir = here / pack_name

# Get the version from the relevant file
version = run_path(pack_name + "/_version.py")
__version__ = version["__version__"]

install_requires = ["transonic", "numpy", "matplotlib"]

relative_paths = ["util.py", "calcul.py"]
make_backend_files([pack_dir / path for path in relative_paths])

extensions = []
if "egg_info" not in sys.argv:
    compile_arch = os.getenv("CARCH", "native")
    extensions = init_transonic_extensions(
        pack_name,
        include_dirs=[np.get_include()],
        compile_args=("-O3", f"-march={compile_arch}", "-DUSE_XSIMD"),
    )

setup(
    version=__version__,
    packages=find_packages(exclude=["doc"]),
    install_requires=install_requires,
    ext_modules=extensions,
)



import numpy as np

from transonic import boost, Array


@boost(wraparound=False, cdivision=True, nonecheck=False)
def moments_hu(nu: "float64[:,:]"):
    hu: Array[np.float64, "1d", "C"] = np.zeros((7,), dtype=np.float64)
    t0: np.float64 = nu[3, 0] + nu[1, 2]
    t1: np.float64 = nu[2, 1] + nu[0, 3]
    q0: np.float64 = t0 * t0
    q1: np.float64 = t1 * t1
    n4: np.float64 = 4 * nu[1, 1]
    s: np.float64 = nu[2, 0] + nu[0, 2]
    d: np.float64 = nu[2, 0] - nu[0, 2]
    hu[0] = s
    hu[1] = d * d + n4 * nu[1, 1]
    hu[3] = q0 + q1
    hu[5] = d * (q0 - q1) + n4 * t0 * t1
    t0 *= q0 - 3 * q1
    t1 *= 3 * q0 - q1
    q0 = nu[3, 0] - 3 * nu[1, 2]
    q1 = 3 * nu[2, 1] - nu[0, 3]
    hu[2] = q0 * q0 + q1 * q1
    hu[4] = q0 * t0 + q1 * t1
    hu[6] = q1 * t0 - q0 * t1
    return np.asarray(hu)



from pathlib import Path

import gast as ast
from transonic.analyses import beniget


from transonic.analyses.util import (
    get_annotations,
    print_ast,
    print_dump,
    print_unparsed,
    filter_code_typevars
)


path_examples = Path("examples")

files = sorted(path_examples.glob("*.py"))

with open(files[1]) as file:
    code = file.read()



print("ast.parse")
module = ast.parse(code)

print("compute DefUseChains")
duc = beniget.DefUseChains()
duc.visit(module)

print("compute Ancestors")
ancestors = beniget.Ancestors()
ancestors.visit(module)


code_filtered = filter_code_typevars(module, duc, ancestors)

print(code_filtered)


from numpy import pi

from transonic import boost

A = "float64[:]"


@boost
def unwrap_1d(image: A, unwrapped_image: A):
    """Phase unwrapping using the naive approach."""
    unwrapped_image[0] = image[0]
    periods = 0
    for i in range(1, image.shape[0]):
        difference = image[i] - image[i - 1]
        if difference > pi:
            periods -= 1
        elif difference < -pi:
            periods += 1
        unwrapped_image[i] = image[i] + 2 * pi * periods



from transonic import jit

# from .base import identity
# FIXME: The above import works but creating an alias does not
from .base import identity as base_identity


identity_jit = jit(native=True)(base_identity)



from transonic import jit, wait_for_all_extensions


@jit
def func_dict(d: "str: float dict"):
    return d.popitem()


d = {"a": 0, "b": 1}
result = func_dict(d)

wait_for_all_extensions()

d1 = {"a": 0, "b": 1}
result1 = func_dict(d1)

assert d == d1
assert result == result1



import numpy as np

from transonic.util import print_versions, timeit_verbose

from mod_pythran import fxfy as fxfy_pythran, fxfy_loops as fxfy_loops_pythran


def fxfy(ft, fn, theta):
    sin_theta = np.sin(theta)
    cos_theta = np.cos(theta)
    fx = (cos_theta * ft) - (sin_theta * fn)
    fy = (sin_theta * ft) + (cos_theta * fn)
    return fx, fy


theta = np.linspace(0, 2 * np.pi, 10000)
ft = 2.5 * theta
fv = 1.5 * theta

print_versions()

loc = locals()

norm = timeit_verbose("fxfy(ft, fv, theta)", globals=loc)
timeit_verbose("fxfy_pythran(ft, fv, theta)", globals=loc, norm=norm)
timeit_verbose("fxfy_loops_pythran(ft, fv, theta)", globals=loc, norm=norm)



import cython

import numpy as np
cimport numpy as np

# not supported, see: https://github.com/cython/cython/issues/754
ctypedef np.ndarray[np.float64_t, ndim=1] A

def mysum(A arr):
    cdef np.float64_t ret = arr.dtype.type(0)
    cdef cython.int i
    for i in range(arr.shape[0]):
        ret += arr[i]
    return ret


import gast
import numpy as np
from exterior_import_boost_2 import func_import_2


const = 1
foo = 1


def func_import():
    return const + func_import_2() + np.pi - np.pi


def use_gast():
    print(gast)



import numpy as np

from transonic import jit


@jit(native=True, xsimd=True)
def row_sum(arr, columns):
    return arr.T[columns].sum(0)


@jit(native=True, xsimd=True)
def row_sum_loops(arr, columns):
    res = np.empty(arr.shape[0], dtype=arr.dtype)
    for i in range(arr.shape[0]):
        sum_ = 0
        for j in range(columns.shape[0]):
            sum_ += arr[i, columns[j]]
        res[i] = sum_
    return res


if __name__ == "__main__":

    from transonic import wait_for_all_extensions

    from util import check, bench

    functions = [row_sum, row_sum_loops]
    arr = np.arange(1_000_000).reshape(1_000, 1_000)
    columns = np.arange(1, 1000, 2)

    check(functions, arr, columns)
    wait_for_all_extensions()
    check(functions, arr, columns)
    bench(functions, arr, columns)



try:
    import cython
except ImportError:
    from transonic_cl import cython


def add(a, b):
    return a + b


def use_add(n=10000):
    tmp = 0
    for _ in range(n):
        tmp = add(tmp, 1)
    return tmp


__transonic__ = ("0.3.3",)



from subprocess import getoutput
import sys

decorator = "boost"
if "jit" in sys.argv:
    decorator = "jit"

print(f"With decorator {decorator}:")


def get_times(backend):
    output = getoutput(
        f"TRANSONIC_BACKEND='{backend}' python row_sum_{decorator}.py"
    )
    lines = output.split("\n")
    index_backend = 1
    if decorator == "jit":
        index_backend = 2
    assert lines[index_backend] == backend.capitalize()
    time_high = float(lines[index_backend + 1].split()[1])
    time_low = float(lines[index_backend + 2].split()[1])
    return time_high, time_low


backend = "python"
time_high, time_low = get_times(backend)

time_ref = time_high


def print_result(backend, time_high, time_low):
    print(backend.capitalize())
    print(f"high level: {time_high:.2e} s  (= {time_high/time_ref:5.2f} * norm)")
    print(f"low level:  {time_low:.2e} s  (= {time_low/time_ref:5.2f} * norm)\n")


print_result(backend, time_high, time_low)

for backend in ("cython", "numba", "pythran"):
    time_high, time_low = get_times(backend)
    print_result(backend, time_high, time_low)



import cython

import numpy as np
cimport numpy as np

ctypedef fused __block0__Array_TypeIfloat_complexI_NDimI1_2Ip1:
   np.ndarray[np.complex_t, ndim=2]
   np.ndarray[np.complex_t, ndim=3]
   np.ndarray[np.float_t, ndim=2]
   np.ndarray[np.float_t, ndim=3]

ctypedef fused __block0__Array_TypeIfloat_complexI_NDimI1_2I:
   np.ndarray[np.complex_t, ndim=1]
   np.ndarray[np.complex_t, ndim=2]
   np.ndarray[np.float_t, ndim=1]
   np.ndarray[np.float_t, ndim=2]

cpdef block0(__block0__Array_TypeIfloat_complexI_NDimI1_2I a, __block0__Array_TypeIfloat_complexI_NDimI1_2Ip1 b, cython.int n)



try:
    import cython
except ImportError:
    from transonic_cl import cython


def block0(a, b, n):

    # transonic block (
    #     A a; A1 b;
    #     int n
    # )
    # transonic block (
    #     int[:] a, b;
    #     float n
    # )
    result = a**2 + b.mean() ** 3 + n
    return result


arguments_blocks = {"block0": ["a", "b", "n"]}

__transonic__ = ("0.3.3",)



import numpy as np
from future._greyreconstruct import reconstruction_loop

from skimage.filters._rank_order import rank_order

y, x = np.mgrid[:20:0.5, :20:0.5]
bumps = np.sin(x) + np.sin(y)
h = 0.3
seed = bumps - h
mask = bumps

assert tuple(seed.shape) == tuple(mask.shape)

selem = np.ones([3] * seed.ndim, dtype=bool)
offset = np.array([d // 2 for d in selem.shape])

# Cross out the center of the selem
selem[tuple(slice(d, d + 1) for d in offset)] = False

# Make padding for edges of reconstructed image so we can ignore boundaries
dims = np.zeros(seed.ndim + 1, dtype=int)
dims[1:] = np.array(seed.shape) + (np.array(selem.shape) - 1)
dims[0] = 2
inside_slices = tuple(slice(o, o + s) for o, s in zip(offset, seed.shape))
# Set padded region to minimum image intensity and mask along first axis so
# we can interleave image and mask pixels when sorting.
pad_value = np.min(seed)

images = np.full(dims, pad_value, dtype="float64")
images[(0, *inside_slices)] = seed
images[(1, *inside_slices)] = mask

# Create a list of strides across the array to get the neighbors within
# a flattened array
value_stride = np.array(images.strides[1:]) // images.dtype.itemsize
image_stride = np.int64(images.strides[0] // images.dtype.itemsize)
selem_mgrid = np.mgrid[[slice(-o, d - o) for d, o in zip(selem.shape, offset)]]
selem_offsets = selem_mgrid[:, selem].transpose()
nb_strides = np.array(
    [np.sum(value_stride * selem_offset) for selem_offset in selem_offsets],
    np.int32,
)

images = images.flatten()

# Erosion goes smallest to largest; dilation goes largest to smallest.
index_sorted = np.argsort(images).astype(np.int32)
index_sorted = index_sorted[::-1]

# Make a linked list of pixels sorted by value. -1 is the list terminator.
prev = np.full(len(images), -1, np.int32)
next_ = np.full(len(images), -1, np.int32)
prev[index_sorted[1:]] = index_sorted[:-1]
next_[index_sorted[:-1]] = index_sorted[1:]

# Cython inner-loop compares the rank of pixel values.
value_rank, value_map = rank_order(images)

start = index_sorted[0]
ranks = np.array(value_rank)
strides = nb_strides
current_idx = np.int64(start)



from transonic import boost

@boost
class MyClass:

    a : int

    def __init__(self, a):
        self.a = a
    @boost
    def method(self, b : int):
        return self.a + b
    
    @boost
    def method2(self, b : int):
        return self.a + b
@boost
class MyClass2:

    a : int

    def __init__(self, a):
        self.a = a
    @boost
    def method(self, b : int):
        o : int
        o = 1
        return self.a + b + o

    @boost
    def method2(self, b : int):
        return self.a + b

    def method3(self, b : int):
        return self.a + b


# cython: language_level=3

import cython

# @cython.ccall
@cython.cfunc
@cython.inline
@cython.returns(cython.int)
@cython.locals(a=cython.int, b=cython.int)
@cython.nogil
def add(a, b):
    return a + b


@cython.ccall
@cython.locals(n=cython.int, i=Py_ssize_t, result=cython.int)
def use_add(n):
    result = 1
    with cython.nogil:
        for i in range(n):
            result = add(result, result)
    return result


from transonic import boost

T = int


@boost(inline=True)
def add(a: T, b: T) -> T:
    return a + b


@boost
def use_add(n: int = 10000):
    tmp: T = 0
    _: T
    for _ in range(n):
        tmp = add(tmp, 1)
    return tmp



from subprocess import getoutput
from pathlib import Path

from transonic.util import timeit

statements = {
    ("cmorph", "_dilate"): "_dilate(image, selem, out, shift_x, shift_y)",
    (
        "_greyreconstruct",
        "reconstruction_loop",
    ): "reconstruction_loop(ranks, prev, next_, strides, current_idx, image_stride)",
}

import_from_skimage = {
    (
        "_greyreconstruct",
        "reconstruction_loop",
    ): "from skimage.morphology._greyreconstruct import reconstruction_loop"
}


def bench_one(name_module="cmorph", func=None, total_duration=2):

    if func is not None:
        raise NotImplementedError

    functions = [
        (mod, func_) for (mod, func_) in statements.keys() if mod == name_module
    ]

    if not functions:
        raise ValueError(f"bad name_module: {name_module}")

    name_function = functions[0][1]

    print(f"module: {name_module}")
    stmt = statements[(name_module, name_function)]
    print(stmt)

    path_setup = Path("setup_codes") / f"{name_module}_{name_function}.py"

    if not path_setup.exists():
        raise RuntimeError

    with open(path_setup) as file:
        setup = file.read()

    if (name_module, name_function) in import_from_skimage:
        setup_from_skimage = setup.replace(
            f"from future.{name_module} import {name_function}",
            import_from_skimage[(name_module, name_function)],
        )
        time = timeit(stmt, setup_from_skimage, total_duration=total_duration)
        print(f"{'from skimage':18s} {time:.2e} s")

    setup_pyx = setup.replace(
        f"from future.{name_module} import", f"from pyx.{name_module} import"
    )

    code = f"""
from transonic.util import timeit
setup = '''{setup}'''
stmt = '''{stmt}'''
print(timeit(stmt, setup, total_duration={total_duration}))
    """

    time_old = timeit(stmt, setup_pyx, total_duration=total_duration)

    print(f"cython pyx skimage {time_old:.2e} s  (= norm)")

    with open("tmp.py", "w") as file:
        file.write(code)

    for backend in ("cython", "pythran", "numba"):
        time = float(getoutput(f"TRANSONIC_BACKEND='{backend}' python tmp.py"))
        print(f"{backend:18s} {time:.2e} s  (= {time/time_old:.2f} * norm)")

    # print(getoutput("TRANSONIC_NO_REPLACE=1 python tmp.py"))

    if (name_module, name_function) not in import_from_skimage:
        return

    setup_from_skimage = setup.replace(
        f"from future.{name_module} import {name_function}",
        import_from_skimage[(name_module, name_function)],
    )
    time = timeit(stmt, setup_from_skimage, total_duration=total_duration)

    print(f"{'from skimage':18s} {time:.2e} s  (= {time/time_old:.2f} * norm)")



import numpy as np


def block0(a, b, n):

    # foo
    # transonic block (
    #     float[][] a, b;
    #     int n
    # ) bar
    # foo
    # transonic block (
    #     float[][][] a, b;
    #     int n
    # )
    # foobar
    result = np.zeros_like(a)
    for _ in range(n):
        result += a**2 + b**3
    return result


def block1(a, b, n):

    # transonic block (
    #     float[][] a, b;
    #     int n
    # )
    # transonic block (
    #     float[][][] a, b;
    #     int n
    # )
    result = np.zeros_like(a)
    for _ in range(n):
        result += a**2 + b**3
    return result


arguments_blocks = {"block0": ["a", "b", "n"], "block1": ["a", "b", "n"]}

__transonic__ = ("0.3.0.post0",)



# __protected__ from numba import njit
# __protected__ @njit(cache=True, fastmath=True)


def __for_method__Myclass__func(self_attr, self_attr2, arg):
    if __for_method__Myclass__func(self_attr, self_attr2, (arg - 1)) < 1:
        return 1
    else:
        a = __for_method__Myclass__func(
            self_attr, self_attr2, (arg - 1)
        ) * __for_method__Myclass__func(self_attr, self_attr2, (arg - 1))
        return (a + ((self_attr * self_attr2) * arg)) + __for_method__Myclass__func(
            self_attr, self_attr2, (arg - 1)
        )


__code_new_method__Myclass__func = "\n\ndef new_method(self, arg):\n    return backend_func(self.attr, self.attr2, arg)\n\n"
__transonic__ = ("0.4.7",)



import cython

import numpy as np
cimport numpy as np

cpdef block0(np.ndarray[np.float_t, ndim=2] a, np.ndarray[np.float_t, ndim=2] b, cython.int n)

cpdef block1(np.ndarray[np.float_t, ndim=2] a, np.ndarray[np.float_t, ndim=2] b, cython.int n)



import cython

import numpy as np
cimport numpy as np

ctypedef fused __compute__Array_TypeIint_complex128I_NDimI1_3I:
   np.ndarray[np.complex128_t, ndim=1]
   np.ndarray[np.complex128_t, ndim=3]
   np.ndarray[np.int_t, ndim=1]
   np.ndarray[np.int_t, ndim=3]

ctypedef fused __compute__TypeIint_complex128I:
   cython.int
   np.complex128_t

ctypedef fused __compute__UnionArray_TypeIint_complex128I_NDimI1_3I_Array_float32_NDimI1_3Ip1:
   np.ndarray[np.complex128_t, ndim=1]
   np.ndarray[np.complex128_t, ndim=3]
   np.ndarray[np.float32_t, ndim=2]
   np.ndarray[np.float32_t, ndim=4]
   np.ndarray[np.int_t, ndim=1]
   np.ndarray[np.int_t, ndim=3]

cpdef compute(__compute__Array_TypeIint_complex128I_NDimI1_3I a, __compute__Array_TypeIint_complex128I_NDimI1_3I b, __compute__TypeIint_complex128I c, __compute__UnionArray_TypeIint_complex128I_NDimI1_3I_Array_float32_NDimI1_3Ip1 d, cython.str e)



import os
import sys
from pathlib import Path
import runpy
import shutil
from contextlib import suppress

import pytest

from transonic.dist import make_backend_files
from transonic.mpi import nb_proc
from transonic.path_data_tests import path_data_tests
from transonic.config import backend_default

cwd = Path.cwd().absolute()
setup_dir = path_data_tests / "test_packaging"


def setup_module():
    os.chdir(setup_dir)
    transonic_src_paths = [setup_dir / "add.py"]
    make_backend_files(transonic_src_paths)


@pytest.fixture
def test_packaging():
    sys.path.append(str(path_data_tests))
    yield "test_packaging"
    sys.path.remove(str(path_data_tests))


@pytest.mark.skipif(backend_default != "pythran", reason="Speedup tests")
@pytest.mark.skipif(not path_data_tests.exists(), reason="no data tests")
@pytest.mark.skipif(nb_proc > 1, reason="No build_ext in MPI")
def test_buildext():
    os.chdir(setup_dir)
    runpy.run_path(str(setup_dir / "setup.py"))


@pytest.mark.xfail(reason="Issue 23")
def test_jit_mod_import(test_packaging):
    """JIT a function from an imported module"""
    runpy.run_module(f"{test_packaging}.base_mod_import")


@pytest.mark.xfail(reason="Issue 23")
def test_jit_func_import(test_packaging):
    """JIT an imported function"""
    runpy.run_module(f"{test_packaging}.base_func_import")


def teardown_module():
    os.chdir(cwd)
    for namedir in ("build", f"__{backend_default}__", "__pycache__"):
        with suppress(FileNotFoundError):
            shutil.rmtree(setup_dir / namedir)

    to_remove = list(setup_dir.glob("*.h")) + list(setup_dir.glob("*.so"))
    for path in to_remove:
        os.remove(path)



import argparse
from bench_util import bench_one

parser = argparse.ArgumentParser(description="Run one benchmark")

parser.add_argument("module", help="Module name")

args = parser.parse_args()

bench_one(args.module)



import numpy as np
from local_module import multiply

from transonic import boost

# transonic def func(float[][], float[][])
# transonic def func(int[][], float[][])


def my_log(b):
    return np.log(b)


@boost
def func(a, b):
    c = multiply(a,b)
    return (c * my_log(b)).max()



"""
See https://foss.heptapod.net/fluiddyn/transonic/issues/6

"""
from transonic import with_blocks, block

T = int


def non_pythranizable(arg):
    """represent a function that can not be compiled by Pythran"""
    return arg


@with_blocks
def func0(arg: int):

    a = b = non_pythranizable(arg)
    c: float = 1.2

    with block():
        # transonic signature(
        #     T a, b
        # )

        # transonic signature(
        a: int
        b: float
        # )

        return a + b + c + arg


@with_blocks
def func1(arg: int):

    a = non_pythranizable(arg)

    with block():
        a: int
        tmp = a + arg

    return non_pythranizable(tmp)



import cython

import numpy as np
cimport numpy as np

ctypedef fused T0:
   np.complex128_t
   np.float64_t

def mysum(np.ndarray[T0, ndim=1] arr):
    cdef T0 ret = arr.dtype.type(0)
    cdef cython.int i
    for i in range(arr.shape[0]):
        ret += arr[i]
    return ret


import cython

import numpy as np
cimport numpy as np

cpdef test_np_fft(np.ndarray[np.float_t, ndim=1] u)

cpdef test_np_linalg_random(np.ndarray[np.float_t, ndim=2] u)

cpdef test_sp_special(cython.int v, cython.float x)



# __protected__ from numba import njit
# __protected__ @njit(cache=True, fastmath=True)


def compute(a, b, c, d, e):
    print(e)
    tmp = a + b
    if 1 and 2:
        tmp *= 2
    return tmp


__transonic__ = ("0.4.7",)



import numpy as np

from transonic import jit, boost

from mod1 import func_import

@boost
class MyClass2:
    def __init__(self):
        self.attr0 = self.attr1 = 1

    @jit()
    def myfunc(self, arg):
        return self.attr1 + self.attr0 + np.abs(arg) + func_import()

    def check(self):
        assert self.myfunc(1) == 4



from transonic import jit, wait_for_all_extensions


@jit
def fib(n: int):
    """fibonacci"""
    if n < 2:
        return n
    return fib(n - 1) + fib(n - 2)


@jit
def use_fib():
    return [fib(n) for n in [1, 3, 5]]


fib2 = fib(2)
result = use_fib()
wait_for_all_extensions()
assert fib2 == fib(2)
assert result == use_fib()



import numpy as np
from transonic.aheadoftime import modules_backends

from numba.targets.registry import CPUDispatcher

from package_cythran.calcul import laplace
from package_cythran.util import func
from package_cythran.other import func_numba

laplace(np.ones((2, 2), dtype=np.int32))
func(1)
func(2.)

ts = modules_backends["pythran"]["package_cythran.calcul"]
assert ts.backend.name == "pythran"

ts = modules_backends["cython"]["package_cythran.util"]
assert ts.backend.name == "cython"

ts = modules_backends["numba"]["package_cythran.other"]
assert ts.backend.name == "numba"

assert isinstance(func_numba, CPUDispatcher)

print("everything looks alright!")



from transonic import jit
import numba

from pure_numpy import laplace_numpy, laplace_loops

laplace_transonic_pythran = jit(native=True, xsimd=True)(laplace_numpy)
laplace_transonic_python = jit(backend="python")(laplace_numpy)
laplace_transonic_numba = jit(backend="numba")(laplace_numpy)
laplace_numba = numba.njit(laplace_numpy)

laplace_transonic_pythran_loops = jit(native=True, xsimd=True)(laplace_loops)
laplace_transonic_python_loops = jit(backend="python")(laplace_loops)
laplace_transonic_numba_loops = jit(backend="numba")(laplace_loops)
laplace_numba_loops = numba.njit(laplace_loops)

if __name__ == "__main__":
    from transonic import wait_for_all_extensions

    from skimage.data import astronaut
    from skimage.color import rgb2gray

    image = astronaut()
    image = rgb2gray(image)

    # warm the functions
    laplace_transonic_python(image)
    laplace_transonic_pythran(image)
    laplace_transonic_pythran_loops(image)
    laplace_transonic_numba(image)
    laplace_transonic_numba_loops(image)
    laplace_numba(image)
    laplace_numba_loops(image)

    wait_for_all_extensions()

    # again warming
    laplace_transonic_numba(image)
    laplace_transonic_numba_loops(image)

    from transonic.util import timeit
    from transonic import __version__
    import pythran

    loc = locals()

    def bench(call, norm=None):
        ret = result = timeit(call, globals=loc)
        if norm is None:
            norm = result
        result /= norm
        print(f"{call.split('(')[0]:33s}: {result:.2f}")
        return ret

    print(
        f"transonic {__version__}\n"
        f"pythran {pythran.__version__}\n"
        f"numba {numba.__version__}\n"
    )

    norm = bench("laplace_transonic_pythran(image)")
    print(f"norm = {norm:.2e} s")
    bench("laplace_transonic_pythran_loops(image)", norm=norm)
    bench("laplace_numba(image)", norm=norm)
    bench("laplace_transonic_numba(image)", norm=norm)
    bench("laplace_numba_loops(image)", norm=norm)
    bench("laplace_transonic_numba_loops(image)", norm=norm)
    bench("laplace_numpy(image)", norm=norm)
    bench("laplace_transonic_python(image)", norm=norm)



# cython: language_level=3
import cython

cpdef inline int add(int a, int b) nogil:
    return a + b

cpdef use_add(cython.int n):
    cdef int i
    cdef int result = 1
    with cython.nogil:
        for i in range(n):
            result = add(result, result)
    return result


# __protected__ from numba import njit
# __protected__ @njit(cache=True, fastmath=True)


def block0(a, b, n):
    # transonic block (
    #     A a; A1 b;
    #     int n
    # )
    # transonic block (
    #     int[:] a, b;
    #     float n
    # )
    result = ((a**2) + (b.mean() ** 3)) + n
    return result


arguments_blocks = {"block0": ["a", "b", "n"]}
__transonic__ = ("0.4.7",)



try:
    import cython
except ImportError:
    from transonic_cl import cython

import numpy as np


def func(a, b):
    return (a * np.log(b)).max()


def func1(a, b):
    return a * np.cos(b)


__transonic__ = ("0.3.3",)



import numpy as np

from transonic import jit


def func0(a, b):
    return a + b


@jit
def func1(a: int, b: int):
    print("b", b)
    return np.exp(a) * b * func0(a, b)


if __name__ == "__main__":

    from time import sleep

    a = b = np.zeros([2, 3])

    for i in range(20):
        print(f"{i}, call with arrays")
        func1(a, b)
        print(f"{i}, call with numbers")
        func1(1, 1.5)
        sleep(1)



try:
    import cython
except ImportError:
    from transonic_cl import cython

from numpy.fft import rfft
from numpy.random import randn
from numpy.linalg import matrix_power
from scipy.special import jv


def test_np_fft(u):
    u_fft = rfft(u)
    return u_fft


def test_np_linalg_random(u):
    (nx, ny) = u.shape
    u[:] = randn(nx, ny)
    u2 = u.T * u
    u4 = matrix_power(u2, 2)
    return u4


def test_sp_special(v, x):
    return jv(v, x)


__transonic__ = ("0.4.2",)



import numpy as np

from transonic import boost, Type, Array, NDim, set_backend_for_this_module

set_backend_for_this_module("pythran")

T = Type(np.int32, np.float64, np.float32)
A = Array[T, NDim(2)]


@boost
def laplace(image: A):
    """Laplace operator in NumPy for 2D images."""
    laplacian = (
        image[:-2, 1:-1]
        + image[2:, 1:-1]
        + image[1:-1, :-2]
        + image[1:-1, 2:]
        - 4 * image[1:-1, 1:-1]
    )
    thresh = np.abs(laplacian) > 0.05
    return thresh



import numpy as np

from transonic import boost, Array

A = Array[float, "3d"]
A1 = Array[float, "1d"]


def expr(a, b):
    return np.arctan2(2 * np.exp(a) ** 2 + 4 * np.log(a * b) ** 3, 2 / a)


def broadcast(a: A, b: A1, out: A):
    out[:] = expr(a, b)


def broadcast_loops(a: A, b: A1, out: A):
    n0, n1, n2 = a.shape
    for i0 in range(n0):
        for i1 in range(n1):
            for i2 in range(n2):
                out[i0, i1, i2] = expr(a[i0, i1, i2], b[i2])


broadcast_pythran = boost(backend="pythran")(broadcast)
broadcast_numba = boost(backend="numba")(broadcast)

broadcast_loops_pythran = boost(backend="pythran")(broadcast_loops)
broadcast_loops_numba = boost(backend="numba")(broadcast_loops)


if __name__ == "__main__":

    from transonic.util import print_versions, timeit_verbose

    print_versions()

    shape = (4, 4, 64)
    a = np.linspace(1, 100, np.prod(shape)).reshape(shape)
    b = np.linspace(1, 100, shape[-1])
    out = np.empty_like(a)
    loc = locals()

    broadcast(a, b, out)
    out_loops = np.empty_like(a)
    broadcast_loops(a, b, out_loops)
    assert np.allclose(out, out_loops)

    print()
    norm = timeit_verbose("broadcast(a, b, out)", globals=loc)

    for backend in ("numba", "pythran"):
        timeit_verbose(f"broadcast_{backend}(a, b, out)", globals=loc, norm=norm)
        timeit_verbose(f"broadcast_loops_{backend}(a, b, out)", globals=loc, norm=norm)



import os
import sys
from distutils.core import setup
from pathlib import Path

import numpy as np

from transonic.dist import make_backend_files, init_transonic_extensions

path_here = Path(__file__).parent.absolute()
include_dirs = [np.get_include()]

pack_name = "future"

paths = tuple((path_here / pack_name).glob("*.py"))

for backend in ("pythran", "cython", "numba"):
    make_backend_files(paths, backend=backend)

extensions = []
if "egg_info" not in sys.argv:
    extensions = init_transonic_extensions(
        pack_name,
        backend="pythran",
        include_dirs=[np.get_include()],
        compile_args=("-O3", "-DUSE_XSIMD"),
        inplace=True,
    )
    extensions.extend(
        init_transonic_extensions(
            pack_name, backend="cython", inplace=True, annotate=True
        )
    )
    init_transonic_extensions(pack_name, backend="numba")


setup(
    name=pack_name,
    ext_modules=extensions,
    # script_name="setup.py",
    script_args=["build_ext", "--inplace"],
    # cmdclass=dict(build_ext=ParallelBuildExt),
)



import numpy as np

from transonic import boost, Array


@boost(wraparound=False, boundscheck=False, cdivision=True, nonecheck=False)
def possible_hull(img: "uint8[:,:]"):
    """Return positions of pixels that possibly belong to the convex hull.

    Parameters
    ----------
    img : ndarray of bool
        Binary input image.

    Returns
    -------
    coords : ndarray (cols, 2)
       The ``(row, column)`` coordinates of all pixels that possibly belong to
       the convex hull.

    """
    r: np.intp
    c: np.intp
    rows: np.intp = img.shape[0]
    cols: np.intp = img.shape[1]

    # Output: rows storage slots for left boundary pixels
    #         cols storage slots for top boundary pixels
    #         rows storage slots for right boundary pixels
    #         cols storage slots for bottom boundary pixels
    coords = np.ones((2 * (rows + cols), 2), dtype=np.intp)
    coords *= -1

    nonzero: Array[np.intp, "2d", "C", "memview"] = coords

    rows_cols: np.intp = rows + cols
    rows_2_cols: np.intp = 2 * rows + cols

    for r in range(rows):

        rows_cols_r = rows_cols + r

        for c in range(cols):

            if img[r, c] != 0:

                rows_c = rows + c
                rows_2_cols_c = rows_2_cols + c

                # Left check
                if nonzero[r, 1] == -1:
                    nonzero[r, 0] = r
                    nonzero[r, 1] = c

                # Right check
                elif nonzero[rows_cols_r, 1] < c:
                    nonzero[rows_cols_r, 0] = r
                    nonzero[rows_cols_r, 1] = c

                # Top check
                if nonzero[rows_c, 1] == -1:
                    nonzero[rows_c, 0] = r
                    nonzero[rows_c, 1] = c

                # Bottom check
                elif nonzero[rows_2_cols_c, 0] < r:
                    nonzero[rows_2_cols_c, 0] = r
                    nonzero[rows_2_cols_c, 1] = c

    return coords[coords[:, 0] != -1]



from transonic import boost


@boost
class Myclass:

    attr: int
    attr2: int

    def __init__(self):
        self.attr = 1
        self.attr2 = 2

    @boost
    def func(self, arg: int):
        if self.func(arg - 1) < 1:
            return 1
        else:
            a = self.func(arg - 1) * self.func(arg - 1)
            return a + self.attr * self.attr2 * arg + self.func(arg - 1)



import numpy as np

# pythran export fxfy(float64[:], float64[:], float64[:])


def fxfy(ft, fn, theta):
    sin_theta = np.sin(theta)
    cos_theta = np.cos(theta)
    fx = (cos_theta * ft) - (sin_theta * fn)
    fy = (sin_theta * ft) + (cos_theta * fn)
    return fx, fy


# pythran export fxfy_loops(float64[:], float64[:], float64[:])


def fxfy_loops(ft, fn, theta):
    n0 = theta.size
    fx = np.empty_like(ft)
    fy = np.empty_like(fn)
    for index in range(n0):
        sin_theta = np.sin(theta[index])
        cos_theta = np.cos(theta[index])
        fx[index] = (cos_theta * ft[index]) - (sin_theta * fn[index])
        fy[index] = (sin_theta * ft[index]) + (cos_theta * fn[index])
    return fx, fy



from numpy.fft import rfft
from numpy.linalg import matrix_power
from numpy.random import randn

from scipy.special import jv

from transonic import boost


@boost
def test_np_fft(u: "float[]"):
    u_fft = rfft(u)
    return u_fft


@boost
def test_np_linalg_random(u: "float[:,:]"):
    nx, ny = u.shape
    u[:] = randn(nx, ny)
    u2 = u.T * u
    u4 = matrix_power(u2, 2)
    return u4


@boost
def test_sp_special(v: int, x: float):
    return jv(v, x)



# __protected__ from numba import njit
import numpy as np
from exterior_import_boost import func_import

# __protected__ @njit(cache=True, fastmath=True)


def __for_method__MyClass2__myfunc(self_attr0, self_attr1, arg):
    return ((self_attr1 + self_attr0) + np.abs(arg)) + func_import()


__code_new_method__MyClass2__myfunc = "\n\ndef new_method(self, arg):\n    return backend_func(self.attr0, self.attr1, arg)\n\n"
__transonic__ = ("0.4.7",)



try:
    import cython
except ImportError:
    from transonic_cl import cython

import numpy as np


def __for_method__Transmitter____call__(self_arr, self_freq, inp):
    "My docstring"
    return ((inp * np.exp(((np.arange(len(inp)) * self_freq) * 1j))), self_arr)


__code_new_method__Transmitter____call__ = """

def new_method(self, inp):
    return backend_func(self.arr, self.freq, inp)

"""

__transonic__ = ("0.3.3",)




import gast as ast

from transonic.analyses.capturex import CaptureX
from transonic.analyses import extast
from transonic.analyses import analyse_aot
from transonic.analyses.util import print_dumped, print_unparsed

from transonic.backends import backends

backend = backends["cython"]

path_file = "simple.py"

with open(path_file) as file:
    code = file.read()

boosted_dicts, code_dependance, annotations, blocks, code_ext = analyse_aot(code, path_file)

module = extast.parse(code)

for node in module.body:
    if isinstance(node, ast.FunctionDef):
        fdef = node
        break

# annotations_locals = annotations["__locals__"]

# annot = annotations["functions"][fdef.name]

# fdef.body = []
# fdef.decorator_list = []

# print_unparsed(fdef)

print("\n".join(backend.get_signatures(fdef.name, fdef, annotations)))

# print_dumped("""
# @cython.locals(result=np.float64_t, i=cython.int)
# def mysum(arr_input):
#     pass
# """)



# __protected__ from numba import njit
import numpy as np

# __protected__ @njit(cache=True, fastmath=True)


def func(a, b):
    return (a * np.log(b)).max()


# __protected__ @njit(cache=True, fastmath=True)


def func1(a, b):
    return a * np.cos(b)


__transonic__ = ("0.4.7",)



from transonic import boost, Type
import numpy as np

T = Type(int, float)
T2 = Type(float, complex)

@boost
def func(m : T, n : T2):
    return np.add(1,n)

@boost
def func1(m : T, n : float = 2):
    t : int = 5.5
    return np.add(1,n) + t



import gast
import numpy as np
from .for_test_exterior_import_jit_2 import func_import_2


const = 1
foo = 1


def func_import_multilevel():
    return const + np.pi - np.pi + func_import_2()


def func_import():
    return const + np.pi - np.pi


def func_import2():
    return const + np.pi - np.pi


def use_gast():
    print(gast)



import cython

import numpy as np
cimport numpy as np

cpdef row_sum(np.ndarray[np.int64_t, ndim=2] arr, np.ndarray[np.int32_t, ndim=1] columns)

@cython.locals(i=np.int32_t, j=np.int32_t, sum_=np.int64_t, res=np.int64_t[:])
cpdef row_sum_loops(np.int64_t[:, :] arr, np.int32_t[:] columns)



try:
    import cython
except ImportError:
    from transonic_cl import cython


def compute(a, b, c, d, e):
    print(e)
    tmp = a + b
    if 1 and 2:
        tmp *= 2
    return tmp


__transonic__ = ("0.3.3",)



import numpy as np


def laplace_numpy(image):
    """Laplace operator in NumPy for 2D images."""
    laplacian = (
        image[:-2, 1:-1]
        + image[2:, 1:-1]
        + image[1:-1, :-2]
        + image[1:-1, 2:]
        - 4 * image[1:-1, 1:-1]
    )
    thresh = np.abs(laplacian) > 0.05
    return thresh


def laplace_loops(image):
    """Laplace operator for 2D images."""
    h = image.shape[0]
    w = image.shape[1]
    laplacian = np.empty((h - 2, w - 2), np.uint8)
    for i in range(1, h - 1):
        for j in range(1, w - 1):
            laplacian[i - 1, j - 1] = (
                np.abs(
                    image[i - 1, j]
                    + image[i + 1, j]
                    + image[i, j - 1]
                    + image[i, j + 1]
                    - 4 * image[i, j]
                )
                > 0.05
            )
    return laplacian




import gast as ast
import beniget

mod = ast.parse("""
T = int
def func() -> T:
    return 1
""")

fdef = mod.body[1]
node = fdef.returns

du = beniget.DefUseChains()
du.visit(mod)

du.chains[node]

ud = beniget.UseDefChains(du)

ud.chains[node]



try:
    import cython
except ImportError:
    from transonic_cl import cython

import numpy as np
from exterior_import_boost import func_import


def func(a, b):
    return (a * np.log(b)).max() + func_import()


__transonic__ = ("0.3.3",)



# __protected__ from numba import njit
import numpy as np
from exterior_import_boost import func_import

# __protected__ @njit(cache=True, fastmath=True)


def func(a, b):
    return (a * np.log(b)).max() + func_import()


__transonic__ = ("0.4.7",)



import numpy as np

from transonic import jit, boost
from transonic.mpi import Path


def func0(a):
    return 2 * a


def func():
    return 1


func0_jitted = jit(func0)


@jit
def func1(a: "int[][] or float[]", l: "int list"):
    tmp = np.exp(sum(l))
    result = tmp * a * func0(a) + func()
    return result


@jit
def func_identity(a):
    return a


# weird but done on purpose for a better coverage
Path(__file__).touch()


@jit()
def func_identity(a):
    return a


@jit
def func_dict(d: "str: float dict"):
    return d.popitem()


@jit
def fib(n: int):
    """fibonacci"""
    if n < 2:
        return n
    return fib(n - 1) + fib(n - 2)


@jit
def use_fib():
    return [fib(n) for n in [1, 3, 5]]


@boost
class MyClass:
    def __init__(self):
        self.attr0 = self.attr1 = 1

    @jit()
    def myfunc(self, arg):
        return self.attr1 + self.attr0 + np.abs(arg)

    @jit
    def myfunc1(self):
        return self.attr0

    def check(self):
        assert self.myfunc(1) == 3


# FIXME support multilevel imported jitted function call in a jitted function
from .for_test_exterior_import_jit import foo, func_import, func_import2
from numpy import pi

const = 1

jitted_func_import = jit(func_import)
jitted_func_import2 = jit(func_import2)


@jit
def main(add: int):
    return (
        add
        + foo
        + pi
        - pi
        - jitted_func_import()
        + func_import2()
        + func0_jitted(1 / 2)
        + const
    )


@boost
class MyClass2:
    def __init__(self):
        self.attr0 = self.attr1 = 1

    @jit()
    def myfunc(self, arg):
        return self.attr1 + self.attr0 + np.abs(arg) + func_import()

    def check(self):
        assert self.myfunc(1) == 4



from transonic import boost

Ai2 = "int[:,:]"


@boost
def _brief_loop(
    image: "float64[:,:]",
    descriptors: "uint8[:,:]",
    keypoints: "intp[:,:]",
    pos0: Ai2,
    pos1: Ai2,
):
    for k in range(len(keypoints)):
        kr, kc = keypoints[k]
        for p in range(len(pos0)):
            pr0, pc0 = pos0[p]
            pr1, pc1 = pos1[p]
            descriptors[k, p] = (
                image[kr + pr0, kc + pc0] < image[kr + pr1, kc + pc1]
            )



import numpy as np

from transonic.aheadoftime import modules_backends

from package_simple.calcul import laplace
from package_simple.util import func

laplace(np.ones((2, 2), dtype=np.int32))
func(1)
func(2.)

modules = modules_backends["pythran"]

ts = modules["package_simple.calcul"]
assert ts.backend.name == "pythran"

ts = modules["package_simple.util"]
assert ts.backend.name == "pythran"

print("everything looks alright!")



try:
    import cython
except ImportError:
    from transonic_cl import cython

import numpy as np


def row_sum(arr, columns):
    return arr.T[columns].sum(0)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.noneckeck(False)
def row_sum_loops(arr, columns):
    # locals type annotations are used only for Cython
    # arr.dtype not supported for memoryview
    dtype = type(arr[(0, 0)])
    res = np.empty(arr.shape[0], dtype=dtype)
    for i in range(arr.shape[0]):
        sum_ = dtype(0)
        for j in range(columns.shape[0]):
            sum_ += arr[(i, columns[j])]
        res[i] = sum_
    return res


__transonic__ = ("0.3.3",)



import numpy as np

from transonic import Type, NDim, Array, boost

T = Type(np.float64, np.complex128)
N = NDim(1)
A = Array[T, N]


@boost
def func(a: A):
    i: int
    n: int = a.shape[0]

    for i in range(n):
        a[i] = a[i] + 1.



import numpy as np
from future.cmorph import _dilate

rows = 1024
cols = 1024
srows = 64
scols = 64

image = np.random.randint(0, 255, rows * cols, dtype=np.uint8).reshape(
    (rows, cols)
)
selem = np.random.randint(0, 1, srows * scols, dtype=np.uint8).reshape(
    (srows, scols)
)
out = np.zeros((rows, cols), dtype=np.uint8)
shift_x = np.int8(2)
shift_y = np.int8(2)


# cython: cdivision=True
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False
import numpy as np
cimport numpy as cnp


cdef inline Py_ssize_t _clip(Py_ssize_t x, Py_ssize_t low,
                             Py_ssize_t high) nogil:
    """Clips coordinate between high and low.

    This method was created so that `hessian_det_appx` does not have to make
    a Python call.

    Parameters
    ----------
    x : int
        Coordinate to be clipped.
    low : int
        The lower bound.
    high : int
        The higher bound.

    Returns
    -------
    x : int
        `x` clipped between `high` and `low`.
    """

    if(x > high):
        return high
    if(x < low):
        return low
    return x


cdef inline cnp.double_t _integ(
    cnp.double_t[:, ::1] img, Py_ssize_t r, Py_ssize_t c,
        Py_ssize_t rl, Py_ssize_t cl) nogil:
    """Integrate over the integral image in the given window

    This method was created so that `hessian_det_appx` does not have to make
    a Python call.

    Parameters
    ----------
    img : array
        The integral image over which to integrate.
    r : int
        The row number of the top left corner.
    c : int
        The column number of the top left corner.
    rl : int
        The number of rows over which to integrate.
    cl : int
        The number of columns over which to integrate.

    Returns
    -------
    ans : int
        The integral over the given window.
    """

    r = _clip(r, 0, img.shape[0] - 1)
    c = _clip(c, 0, img.shape[1] - 1)

    r2 = _clip(r + rl, 0, img.shape[0] - 1)
    c2 = _clip(c + cl, 0, img.shape[1] - 1)

    cdef cnp.double_t ans = img[r, c] + img[r2, c2] - img[r, c2] - img[r2, c]

    if (ans < 0):
        return 0
    return ans


def _hessian_matrix_det(cnp.double_t[:, ::1] img, double sigma):
    """Computes the approximate Hessian Determinant over an image.

    This method uses box filters over integral images to compute the
    approximate Hessian Determinant as described in [1]_.

    Parameters
    ----------
    img : array
        The integral image over which to compute Hessian Determinant.
    sigma : float
        Standard deviation used for the Gaussian kernel, used for the Hessian
        matrix

    Returns
    -------
    out : array
        The array of the Determinant of Hessians.

    References
    ----------
    .. [1] Herbert Bay, Andreas Ess, Tinne Tuytelaars, Luc Van Gool,
           "SURF: Speeded Up Robust Features"
           ftp://ftp.vision.ee.ethz.ch/publications/articles/eth_biwi_00517.pdf

    Notes
    -----
    The running time of this method only depends on size of the image. It is
    independent of `sigma` as one would expect. The downside is that the
    result for `sigma` less than `3` is not accurate, i.e., not similar to
    the result obtained if someone computed the Hessian and took it's
    determinant.
    """

    cdef Py_ssize_t size = int(3 * sigma)
    cdef Py_ssize_t height = img.shape[0]
    cdef Py_ssize_t width = img.shape[1]
    cdef Py_ssize_t r, c
    cdef Py_ssize_t s2 = (size - 1) / 2
    cdef Py_ssize_t s3 = size / 3
    cdef Py_ssize_t l = size / 3
    cdef Py_ssize_t w = size
    cdef Py_ssize_t b = (size - 1) / 2
    cdef cnp.double_t mid, side, tl, tr, bl, br
    cdef cnp.double_t[:, ::1] out = np.zeros_like(img, dtype=np.double)
    cdef cnp.double_t w_i = 1.0 / size / size

    cdef float dxx, dyy, dxy

    with nogil:
        if size % 2 == 0:
            size += 1

        for r in range(height):
            for c in range(width):
                tl = _integ(img, r - s3, c - s3, s3, s3)  # top left
                br = _integ(img, r + 1, c + 1, s3, s3)  # bottom right
                bl = _integ(img, r - s3, c + 1, s3, s3)  # bottom left
                tr = _integ(img, r + 1, c - s3, s3, s3)  # top right

                dxy = bl + tr - tl - br
                dxy = -dxy * w_i

                mid = _integ(img, r - s3 + 1, c - s2, 2 * s3 - 1, w)  # middle box
                side = _integ(img, r - s3 + 1, c - s3 / 2, 2 * s3 - 1, s3)  # sides

                dxx = mid - 3 * side
                dxx = -dxx * w_i

                mid = _integ(img, r - s2, c - s3 + 1, w, 2 * s3 - 1)
                side = _integ(img, r - s3 / 2, c - s3 + 1, s3, 2 * s3 - 1)

                dyy = mid - 3 * side
                dyy = -dyy * w_i

                out[r, c] = (dxx * dyy - 0.81 * (dxy * dxy))

    return out



from pathlib import Path

from transonic.analyses.util import print_dumped


from transonic.backends.pythran import make_pythran_code

path_examples = Path("examples")

paths = sorted(path_examples.glob("*.py"))
path = paths[4]

# path = Path("~/Dev/fluidsim/fluidsim/base/time_stepping/pseudo_spect.py")

code = make_pythran_code(path)

print(code)



import numpy as np

from transonic import boost
from exterior_import_boost import func_import


@boost
class MyClass2:
    attr0: int
    attr1: int

    def __init__(self):
        self.attr0 = self.attr1 = 1

    @boost
    def myfunc(self, arg: int):
        return self.attr1 + self.attr0 + np.abs(arg) + func_import()

    def check(self):
        assert self.myfunc(1) == 5



import numpy as np

from transonic import boost, Array

T_index = np.int32
# we use a type variable because it can be replaced by a fused type.
T = np.int64
A1d_i = Array[T_index, "1d"]
A1d = Array[T, "1d"]
A2d = Array[T, "2d"]
V1d_i = Array[T_index, "1d", "memview"]
V1d = Array[T, "1d", "memview"]
V2d = Array[T, "2d", "memview"]


@boost
def row_sum(arr: A2d, columns: A1d_i):
    return arr.T[columns].sum(0)


@boost(boundscheck=False, wraparound=False, cdivision=True, nonecheck=False)
def row_sum_loops(arr: V2d, columns: V1d_i):
    # locals type annotations are used only for Cython
    i: T_index
    j: T_index
    sum_: T
    # arr.dtype not supported for memoryview
    dtype = type(arr[0, 0])
    res: V1d = np.empty(arr.shape[0], dtype=dtype)
    for i in range(arr.shape[0]):
        sum_ = dtype(0)
        for j in range(columns.shape[0]):
            sum_ += arr[i, columns[j]]
        res[i] = sum_
    return res


if __name__ == "__main__":

    from util import check, bench

    functions = [row_sum, row_sum_loops]
    arr = np.arange(1_000_000).reshape(1_000, 1_000)
    columns = np.arange(1, 1000, 2, dtype=T_index)

    check(functions, arr, columns)
    bench(functions, arr, columns)



import gast as ast

from transonic.analyses.util import get_annotations

code = """

type_ = float

def myfunc(a: int, b: type_):
    pass

"""

module = ast.parse(code)
fdef = module.body[1]

print(get_annotations(fdef, {"type_": float}))


#cython: cdivision=True
#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False

cimport numpy as cnp


def _brief_loop(double[:, ::1] image, unsigned char[:, ::1] descriptors,
                Py_ssize_t[:, ::1] keypoints,
                int[:, ::1] pos0, int[:, ::1] pos1):

    cdef Py_ssize_t k, d, kr, kc, pr0, pr1, pc0, pc1

    with nogil:
        for p in range(pos0.shape[0]):
            pr0 = pos0[p, 0]
            pc0 = pos0[p, 1]
            pr1 = pos1[p, 0]
            pc1 = pos1[p, 1]
            for k in range(keypoints.shape[0]):
                kr = keypoints[k, 0]
                kc = keypoints[k, 1]
                if image[kr + pr0, kc + pc0] < image[kr + pr1, kc + pc1]:
                    descriptors[k, p] = True
